@@ -1,6 +1,6 @@
-/* Copyright (C) 2002, 2003, 2004, 2005, 2006 Free Software Foundation, Inc.
+/* Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007 Free Software Foundation, Inc.
    Contributed by Andy Vaught
-   Namelist output contibuted by Paul Thomas
+   Namelist output contributed by Paul Thomas
 
 This file is part of the GNU Fortran 95 runtime library (libgfortran).
 
@@ -28,23 +28,15 @@ along with Libgfortran; see the file COPYING.  If not, write to
 the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 Boston, MA 02110-1301, USA.  */
 
-#include "config.h"
+#include "io.h"
 #include <assert.h>
 #include <string.h>
 #include <ctype.h>
-#include <float.h>
-#include <stdio.h>
 #include <stdlib.h>
-#include "libgfortran.h"
-#include "io.h"
-
+#include <stdbool.h>
 #define star_fill(p, n) memset(p, '*', n)
 
-
-typedef enum
-{ SIGN_NONE, SIGN_MINUS, SIGN_PLUS }
-sign_t;
-
+#include "write_float.def"
 
 void
 write_a (st_parameter_dt *dtp, const fnode *f, const char *source, int len)
@@ -54,17 +46,78 @@ write_a (st_parameter_dt *dtp, const fnode *f, const char *source, int len)
 
   wlen = f->u.string.length < 0 ? len : f->u.string.length;
 
-  p = write_block (dtp, wlen);
-  if (p == NULL)
-    return;
+#ifdef HAVE_CRLF
+  /* If this is formatted STREAM IO convert any embedded line feed characters
+     to CR_LF on systems that use that sequence for newlines.  See F2003
+     Standard sections 10.6.3 and 9.9 for further information.  */
+  if (is_stream_io (dtp))
+    {
+      const char crlf[] = "\r\n";
+      int i, q, bytes;
+      q = bytes = 0;
 
-  if (wlen < len)
-    memcpy (p, source, wlen);
+      /* Write out any padding if needed.  */
+      if (len < wlen)
+	{
+	  p = write_block (dtp, wlen - len);
+	  if (p == NULL)
+	    return;
+	  memset (p, ' ', wlen - len);
+	}
+
+      /* Scan the source string looking for '\n' and convert it if found.  */
+      for (i = 0; i < wlen; i++)
+	{
+	  if (source[i] == '\n')
+	    {
+	      /* Write out the previously scanned characters in the string.  */
+	      if (bytes > 0)
+		{
+		  p = write_block (dtp, bytes);
+		  if (p == NULL)
+		    return;
+		  memcpy (p, &source[q], bytes);
+		  q += bytes;
+		  bytes = 0;
+		}
+
+	      /* Write out the CR_LF sequence.  */ 
+	      q++;
+	      p = write_block (dtp, 2);
+              if (p == NULL)
+                return;
+	      memcpy (p, crlf, 2);
+	    }
+	  else
+	    bytes++;
+	}
+
+      /*  Write out any remaining bytes if no LF was found.  */
+      if (bytes > 0)
+	{
+	  p = write_block (dtp, bytes);
+	  if (p == NULL)
+	    return;
+	  memcpy (p, &source[q], bytes);
+	}
+    }
   else
     {
-      memset (p, ' ', wlen - len);
-      memcpy (p + wlen - len, source, len);
+#endif
+      p = write_block (dtp, wlen);
+      if (p == NULL)
+	return;
+
+      if (wlen < len)
+	memcpy (p, source, wlen);
+      else
+	{
+	  memset (p, ' ', wlen - len);
+	  memcpy (p + wlen - len, source, len);
+	}
+#ifdef HAVE_CRLF
     }
+#endif
 }
 
 static GFC_INTEGER_LARGEST
@@ -175,633 +228,6 @@ extract_uint (const void *p, int len)
   return i;
 }
 
-static GFC_REAL_LARGEST
-extract_real (const void *p, int len)
-{
-  GFC_REAL_LARGEST i = 0;
-  switch (len)
-    {
-    case 4:
-      {
-	GFC_REAL_4 tmp;
-	memcpy ((void *) &tmp, p, len);
-	i = tmp;
-      }
-      break;
-    case 8:
-      {
-	GFC_REAL_8 tmp;
-	memcpy ((void *) &tmp, p, len);
-	i = tmp;
-      }
-      break;
-#ifdef HAVE_GFC_REAL_10
-    case 10:
-      {
-	GFC_REAL_10 tmp;
-	memcpy ((void *) &tmp, p, len);
-	i = tmp;
-      }
-      break;
-#endif
-#ifdef HAVE_GFC_REAL_16
-    case 16:
-      {
-	GFC_REAL_16 tmp;
-	memcpy ((void *) &tmp, p, len);
-	i = tmp;
-      }
-      break;
-#endif
-    default:
-      internal_error (NULL, "bad real kind");
-    }
-  return i;
-}
-
-
-/* Given a flag that indicate if a value is negative or not, return a
-   sign_t that gives the sign that we need to produce.  */
-
-static sign_t
-calculate_sign (st_parameter_dt *dtp, int negative_flag)
-{
-  sign_t s = SIGN_NONE;
-
-  if (negative_flag)
-    s = SIGN_MINUS;
-  else
-    switch (dtp->u.p.sign_status)
-      {
-      case SIGN_SP:
-	s = SIGN_PLUS;
-	break;
-      case SIGN_SS:
-	s = SIGN_NONE;
-	break;
-      case SIGN_S:
-	s = options.optional_plus ? SIGN_PLUS : SIGN_NONE;
-	break;
-      }
-
-  return s;
-}
-
-
-/* Returns the value of 10**d.  */
-
-static GFC_REAL_LARGEST
-calculate_exp (int d)
-{
-  int i;
-  GFC_REAL_LARGEST r = 1.0;
-
-  for (i = 0; i< (d >= 0 ? d : -d); i++)
-    r *= 10;
-
-  r = (d >= 0) ? r : 1.0 / r;
-
-  return r;
-}
-
-
-/* Generate corresponding I/O format for FMT_G output.
-   The rules to translate FMT_G to FMT_E or FMT_F from DEC fortran
-   LRM (table 11-2, Chapter 11, "I/O Formatting", P11-25) is:
-
-   Data Magnitude                              Equivalent Conversion
-   0< m < 0.1-0.5*10**(-d-1)                   Ew.d[Ee]
-   m = 0                                       F(w-n).(d-1), n' '
-   0.1-0.5*10**(-d-1)<= m < 1-0.5*10**(-d)     F(w-n).d, n' '
-   1-0.5*10**(-d)<= m < 10-0.5*10**(-d+1)      F(w-n).(d-1), n' '
-   10-0.5*10**(-d+1)<= m < 100-0.5*10**(-d+2)  F(w-n).(d-2), n' '
-   ................                           ..........
-   10**(d-1)-0.5*10**(-1)<= m <10**d-0.5       F(w-n).0,n(' ')
-   m >= 10**d-0.5                              Ew.d[Ee]
-
-   notes: for Gw.d ,  n' ' means 4 blanks
-          for Gw.dEe, n' ' means e+2 blanks  */
-
-static fnode *
-calculate_G_format (st_parameter_dt *dtp, const fnode *f,
-		    GFC_REAL_LARGEST value, int *num_blank)
-{
-  int e = f->u.real.e;
-  int d = f->u.real.d;
-  int w = f->u.real.w;
-  fnode *newf;
-  GFC_REAL_LARGEST m, exp_d;
-  int low, high, mid;
-  int ubound, lbound;
-
-  newf = get_mem (sizeof (fnode));
-
-  /* Absolute value.  */
-  m = (value > 0.0) ? value : -value;
-
-  /* In case of the two data magnitude ranges,
-     generate E editing, Ew.d[Ee].  */
-  exp_d = calculate_exp (d);
-  if ((m > 0.0 && m < 0.1 - 0.05 / exp_d) || (m >= exp_d - 0.5 ) ||
-      ((m == 0.0) && !(compile_options.allow_std & GFC_STD_F2003)))
-    {
-      newf->format = FMT_E;
-      newf->u.real.w = w;
-      newf->u.real.d = d;
-      newf->u.real.e = e;
-      *num_blank = 0;
-      return newf;
-    }
-
-  /* Use binary search to find the data magnitude range.  */
-  mid = 0;
-  low = 0;
-  high = d + 1;
-  lbound = 0;
-  ubound = d + 1;
-
-  while (low <= high)
-    {
-      GFC_REAL_LARGEST temp;
-      mid = (low + high) / 2;
-
-      /* 0.1 * 10**mid - 0.5 * 10**(mid-d-1)  */
-      temp = 0.1 * calculate_exp (mid) - 0.5 * calculate_exp (mid - d - 1);
-
-      if (m < temp)
-        {
-          ubound = mid;
-          if (ubound == lbound + 1)
-            break;
-          high = mid - 1;
-        }
-      else if (m > temp)
-        {
-          lbound = mid;
-          if (ubound == lbound + 1)
-            {
-              mid ++;
-              break;
-            }
-          low = mid + 1;
-        }
-      else
-        break;
-    }
-
-  /* Pad with blanks where the exponent would be.  */
-  if (e < 0)
-    *num_blank = 4;
-  else
-    *num_blank = e + 2;
-
-  /* Generate the F editing. F(w-n).(-(mid-d-1)), n' '.  */
-  newf->format = FMT_F;
-  newf->u.real.w = f->u.real.w - *num_blank;
-
-  /* Special case.  */
-  if (m == 0.0)
-    newf->u.real.d = d - 1;
-  else
-    newf->u.real.d = - (mid - d - 1);
-
-  /* For F editing, the scale factor is ignored.  */
-  dtp->u.p.scale_factor = 0;
-  return newf;
-}
-
-
-/* Output a real number according to its format which is FMT_G free.  */
-
-static void
-output_float (st_parameter_dt *dtp, const fnode *f, GFC_REAL_LARGEST value)
-{
-#if defined(HAVE_GFC_REAL_16) && __LDBL_DIG__ > 18
-# define MIN_FIELD_WIDTH 46
-#else
-# define MIN_FIELD_WIDTH 31
-#endif
-#define STR(x) STR1(x)
-#define STR1(x) #x
-  /* This must be large enough to accurately hold any value.  */
-  char buffer[MIN_FIELD_WIDTH+1];
-  char *out;
-  char *digits;
-  int e;
-  char expchar;
-  format_token ft;
-  int w;
-  int d;
-  int edigits;
-  int ndigits;
-  /* Number of digits before the decimal point.  */
-  int nbefore;
-  /* Number of zeros after the decimal point.  */
-  int nzero;
-  /* Number of digits after the decimal point.  */
-  int nafter;
-  /* Number of zeros after the decimal point, whatever the precision.  */
-  int nzero_real;
-  int leadzero;
-  int nblanks;
-  int i;
-  sign_t sign;
-  double abslog;
-
-  ft = f->format;
-  w = f->u.real.w;
-  d = f->u.real.d;
-
-  nzero_real = -1;
-
-
-  /* We should always know the field width and precision.  */
-  if (d < 0)
-    internal_error (&dtp->common, "Unspecified precision");
-
-  /* Use sprintf to print the number in the format +D.DDDDe+ddd
-     For an N digit exponent, this gives us (MIN_FIELD_WIDTH-5)-N digits
-     after the decimal point, plus another one before the decimal point.  */
-  sign = calculate_sign (dtp, value < 0.0);
-  if (value < 0)
-    value = -value;
-
-  /* Printf always prints at least two exponent digits.  */
-  if (value == 0)
-    edigits = 2;
-  else
-    {
-#if defined(HAVE_GFC_REAL_10) || defined(HAVE_GFC_REAL_16)
-      abslog = fabs((double) log10l(value));
-#else
-      abslog = fabs(log10(value));
-#endif
-      if (abslog < 100)
-	edigits = 2;
-      else
-        edigits = 1 + (int) log10(abslog);
-    }
-
-  if (ft == FMT_F || ft == FMT_EN
-      || ((ft == FMT_D || ft == FMT_E) && dtp->u.p.scale_factor != 0))
-    {
-      /* Always convert at full precision to avoid double rounding.  */
-      ndigits = MIN_FIELD_WIDTH - 4 - edigits;
-    }
-  else
-    {
-      /* We know the number of digits, so can let printf do the rounding
-	 for us.  */
-      if (ft == FMT_ES)
-	ndigits = d + 1;
-      else
-	ndigits = d;
-      if (ndigits > MIN_FIELD_WIDTH - 4 - edigits)
-	ndigits = MIN_FIELD_WIDTH - 4 - edigits;
-    }
-
-  /* #   The result will always contain a decimal point, even if no
-   *     digits follow it
-   *
-   * -   The converted value is to be left adjusted on the field boundary
-   *
-   * +   A sign (+ or -) always be placed before a number
-   *
-   * MIN_FIELD_WIDTH  minimum field width
-   *
-   * *   (ndigits-1) is used as the precision
-   *
-   *   e format: [-]d.ddde±dd where there is one digit before the
-   *   decimal-point character and the number of digits after it is
-   *   equal to the precision. The exponent always contains at least two
-   *   digits; if the value is zero, the exponent is 00.
-   */
-  sprintf (buffer, "%+-#" STR(MIN_FIELD_WIDTH) ".*"
-	   GFC_REAL_LARGEST_FORMAT "e", ndigits - 1, value);
-
-  /* Check the resulting string has punctuation in the correct places.  */
-  if (d != 0 && (buffer[2] != '.' || buffer[ndigits + 2] != 'e'))
-      internal_error (&dtp->common, "printf is broken");
-
-  /* Read the exponent back in.  */
-  e = atoi (&buffer[ndigits + 3]) + 1;
-
-  /* Make sure zero comes out as 0.0e0.  */
-  if (value == 0.0)
-    e = 0;
-
-  /* Normalize the fractional component.  */
-  buffer[2] = buffer[1];
-  digits = &buffer[2];
-
-  /* Figure out where to place the decimal point.  */
-  switch (ft)
-    {
-    case FMT_F:
-      nbefore = e + dtp->u.p.scale_factor;
-      if (nbefore < 0)
-	{
-	  nzero = -nbefore;
-          nzero_real = nzero;
-	  if (nzero > d)
-	    nzero = d;
-	  nafter = d - nzero;
-	  nbefore = 0;
-	}
-      else
-	{
-	  nzero = 0;
-	  nafter = d;
-	}
-      expchar = 0;
-      break;
-
-    case FMT_E:
-    case FMT_D:
-      i = dtp->u.p.scale_factor;
-      if (value != 0.0)
-	e -= i;
-      if (i < 0)
-	{
-	  nbefore = 0;
-	  nzero = -i;
-	  nafter = d + i;
-	}
-      else if (i > 0)
-	{
-	  nbefore = i;
-	  nzero = 0;
-	  nafter = (d - i) + 1;
-	}
-      else /* i == 0 */
-	{
-	  nbefore = 0;
-	  nzero = 0;
-	  nafter = d;
-	}
-
-      if (ft == FMT_E)
-	expchar = 'E';
-      else
-	expchar = 'D';
-      break;
-
-    case FMT_EN:
-      /* The exponent must be a multiple of three, with 1-3 digits before
-	 the decimal point.  */
-      if (value != 0.0)
-        e--;
-      if (e >= 0)
-	nbefore = e % 3;
-      else
-	{
-	  nbefore = (-e) % 3;
-	  if (nbefore != 0)
-	    nbefore = 3 - nbefore;
-	}
-      e -= nbefore;
-      nbefore++;
-      nzero = 0;
-      nafter = d;
-      expchar = 'E';
-      break;
-
-    case FMT_ES:
-      if (value != 0.0)
-        e--;
-      nbefore = 1;
-      nzero = 0;
-      nafter = d;
-      expchar = 'E';
-      break;
-
-    default:
-      /* Should never happen.  */
-      internal_error (&dtp->common, "Unexpected format token");
-    }
-
-  /* Round the value.  */
-  if (nbefore + nafter == 0)
-    {
-      ndigits = 0;
-      if (nzero_real == d && digits[0] >= '5')
-        {
-          /* We rounded to zero but shouldn't have */
-          nzero--;
-          nafter = 1;
-          digits[0] = '1';
-          ndigits = 1;
-        }
-    }
-  else if (nbefore + nafter < ndigits)
-    {
-      ndigits = nbefore + nafter;
-      i = ndigits;
-      if (digits[i] >= '5')
-	{
-	  /* Propagate the carry.  */
-	  for (i--; i >= 0; i--)
-	    {
-	      if (digits[i] != '9')
-		{
-		  digits[i]++;
-		  break;
-		}
-	      digits[i] = '0';
-	    }
-
-	  if (i < 0)
-	    {
-	      /* The carry overflowed.  Fortunately we have some spare space
-		 at the start of the buffer.  We may discard some digits, but
-		 this is ok because we already know they are zero.  */
-	      digits--;
-	      digits[0] = '1';
-	      if (ft == FMT_F)
-		{
-		  if (nzero > 0)
-		    {
-		      nzero--;
-		      nafter++;
-		    }
-		  else
-		    nbefore++;
-		}
-	      else if (ft == FMT_EN)
-		{
-		  nbefore++;
-		  if (nbefore == 4)
-		    {
-		      nbefore = 1;
-		      e += 3;
-		    }
-		}
-	      else
-		e++;
-	    }
-	}
-    }
-
-  /* Calculate the format of the exponent field.  */
-  if (expchar)
-    {
-      edigits = 1;
-      for (i = abs (e); i >= 10; i /= 10)
-	edigits++;
-
-      if (f->u.real.e < 0)
-	{
-	  /* Width not specified.  Must be no more than 3 digits.  */
-	  if (e > 999 || e < -999)
-	    edigits = -1;
-	  else
-	    {
-	      edigits = 4;
-	      if (e > 99 || e < -99)
-		expchar = ' ';
-	    }
-	}
-      else
-	{
-	  /* Exponent width specified, check it is wide enough.  */
-	  if (edigits > f->u.real.e)
-	    edigits = -1;
-	  else
-	    edigits = f->u.real.e + 2;
-	}
-    }
-  else
-    edigits = 0;
-
-  /* Pick a field size if none was specified.  */
-  if (w <= 0)
-    w = nbefore + nzero + nafter + (sign != SIGN_NONE ? 2 : 1);
-
-  /* Create the ouput buffer.  */
-  out = write_block (dtp, w);
-  if (out == NULL)
-    return;
-
-  /* Zero values always output as positive, even if the value was negative
-     before rounding.  */
-  for (i = 0; i < ndigits; i++)
-    {
-      if (digits[i] != '0')
-	break;
-    }
-  if (i == ndigits)
-    sign = calculate_sign (dtp, 0);
-
-  /* Work out how much padding is needed.  */
-  nblanks = w - (nbefore + nzero + nafter + edigits + 1);
-  if (sign != SIGN_NONE)
-    nblanks--;
-
-  /* Check the value fits in the specified field width.  */
-  if (nblanks < 0 || edigits == -1)
-    {
-      star_fill (out, w);
-      return;
-    }
-
-  /* See if we have space for a zero before the decimal point.  */
-  if (nbefore == 0 && nblanks > 0)
-    {
-      leadzero = 1;
-      nblanks--;
-    }
-  else
-    leadzero = 0;
-
-  /* Pad to full field width.  */
-
-
-  if ( ( nblanks > 0 ) && !dtp->u.p.no_leading_blank)
-    {
-      memset (out, ' ', nblanks);
-      out += nblanks;
-    }
-
-  /* Output the initial sign (if any).  */
-  if (sign == SIGN_PLUS)
-    *(out++) = '+';
-  else if (sign == SIGN_MINUS)
-    *(out++) = '-';
-
-  /* Output an optional leading zero.  */
-  if (leadzero)
-    *(out++) = '0';
-
-  /* Output the part before the decimal point, padding with zeros.  */
-  if (nbefore > 0)
-    {
-      if (nbefore > ndigits)
-	i = ndigits;
-      else
-	i = nbefore;
-
-      memcpy (out, digits, i);
-      while (i < nbefore)
-	out[i++] = '0';
-
-      digits += i;
-      ndigits -= i;
-      out += nbefore;
-    }
-  /* Output the decimal point.  */
-  *(out++) = '.';
-
-  /* Output leading zeros after the decimal point.  */
-  if (nzero > 0)
-    {
-      for (i = 0; i < nzero; i++)
-	*(out++) = '0';
-    }
-
-  /* Output digits after the decimal point, padding with zeros.  */
-  if (nafter > 0)
-    {
-      if (nafter > ndigits)
-	i = ndigits;
-      else
-	i = nafter;
-
-      memcpy (out, digits, i);
-      while (i < nafter)
-	out[i++] = '0';
-
-      digits += i;
-      ndigits -= i;
-      out += nafter;
-    }
-
-  /* Output the exponent.  */
-  if (expchar)
-    {
-      if (expchar != ' ')
-	{
-	  *(out++) = expchar;
-	  edigits--;
-	}
-#if HAVE_SNPRINTF
-      snprintf (buffer, sizeof (buffer), "%+0*d", edigits, e);
-#else
-      sprintf (buffer, "%+0*d", edigits, e);
-#endif
-      memcpy (out, buffer, edigits);
-    }
-
-  if (dtp->u.p.no_leading_blank)
-    {
-      out += edigits;
-      memset( out , ' ' , nblanks );
-      dtp->u.p.no_leading_blank = 0;
-    }
-#undef STR
-#undef STR1
-#undef MIN_FIELD_WIDTH
-}
-
 
 void
 write_l (st_parameter_dt *dtp, const fnode *f, char *source, int len)
@@ -818,108 +244,6 @@ write_l (st_parameter_dt *dtp, const fnode *f, char *source, int len)
   p[f->u.w - 1] = (n) ? 'T' : 'F';
 }
 
-/* Output a real number according to its format.  */
-
-static void
-write_float (st_parameter_dt *dtp, const fnode *f, const char *source, int len)
-{
-  GFC_REAL_LARGEST n;
-  int nb =0, res, save_scale_factor;
-  char * p, fin;
-  fnode *f2 = NULL;
-
-  n = extract_real (source, len);
-
-  if (f->format != FMT_B && f->format != FMT_O && f->format != FMT_Z)
-    {
-      res = isfinite (n); 
-      if (res == 0)
-	{
-	  nb =  f->u.real.w;
-	  
-	  /* If the field width is zero, the processor must select a width 
-	     not zero.  4 is chosen to allow output of '-Inf' or '+Inf' */
-	     
-	  if (nb == 0) nb = 4;
-	  p = write_block (dtp, nb);
-          if (p == NULL)
-            return;
-	  if (nb < 3)
-	    {
-	      memset (p, '*',nb);
-	      return;
-	    }
-
-	  memset(p, ' ', nb);
-	  res = !isnan (n);
-	  if (res != 0)
-	    {
-	      if (signbit(n))
-	        {
-	        
-	          /* If the sign is negative and the width is 3, there is
-	             insufficient room to output '-Inf', so output asterisks */
-	             
-	          if (nb == 3)
-	            {
-	              memset (p, '*',nb);
-	              return;
-	            }
-	            
-	          /* The negative sign is mandatory */
-	            
-	          fin = '-';
-		}    
-	      else
-	      
-	          /* The positive sign is optional, but we output it for
-	             consistency */
-	             
-		  fin = '+';
-
-	      if (nb > 8)
-	      
-	        /* We have room, so output 'Infinity' */
-	        
-		memcpy(p + nb - 8, "Infinity", 8);
-	      else
-	      
-	        /* For the case of width equals 8, there is not enough room
-	           for the sign and 'Infinity' so we go with 'Inf' */
-	            
-		memcpy(p + nb - 3, "Inf", 3);
-	      if (nb < 9 && nb > 3)
-		p[nb - 4] = fin;  /* Put the sign in front of Inf */
-	      else if (nb > 8)
-		p[nb - 9] = fin;  /* Put the sign in front of Infinity */
-	    }
-	  else
-	    memcpy(p + nb - 3, "NaN", 3);
-	  return;
-	}
-    }
-
-  if (f->format != FMT_G)
-    output_float (dtp, f, n);
-  else
-    {
-      save_scale_factor = dtp->u.p.scale_factor;
-      f2 = calculate_G_format (dtp, f, n, &nb);
-      output_float (dtp, f2, n);
-      dtp->u.p.scale_factor = save_scale_factor;
-      if (f2 != NULL)
-        free_mem(f2);
-
-      if (nb > 0)
-        {
-	  p = write_block (dtp, nb);
-          if (p == NULL)
-            return;
-          memset (p, ' ', nb);
-        }
-    }
-}
-
 
 static void
 write_int (st_parameter_dt *dtp, const fnode *f, const char *source, int len,
@@ -1374,23 +698,23 @@ write_real (st_parameter_dt *dtp, const char *source, int length)
   switch (length)
     {
     case 4:
-      f.u.real.w = 14;
-      f.u.real.d = 7;
+      f.u.real.w = 15;
+      f.u.real.d = 8;
       f.u.real.e = 2;
       break;
     case 8:
-      f.u.real.w = 23;
-      f.u.real.d = 15;
+      f.u.real.w = 25;
+      f.u.real.d = 17;
       f.u.real.e = 3;
       break;
     case 10:
-      f.u.real.w = 28;
-      f.u.real.d = 19;
+      f.u.real.w = 29;
+      f.u.real.d = 20;
       f.u.real.e = 4;
       break;
     case 16:
-      f.u.real.w = 43;
-      f.u.real.d = 34;
+      f.u.real.w = 44;
+      f.u.real.d = 35;
       f.u.real.e = 4;
       break;
     default:
@@ -1541,6 +865,10 @@ nml_write_obj (st_parameter_dt *dtp, namelist_info * obj, index_type offset,
   char rep_buff[NML_DIGITS];
   namelist_info * cmp;
   namelist_info * retval = obj->next;
+  size_t base_name_len;
+  size_t base_var_name_len;
+  size_t tot_len;
+  unit_delim tmp_delim;
 
   /* Write namelist variable names in upper case. If a derived type,
      nothing is output.  If a component, base and base_name are set.  */
@@ -1636,7 +964,7 @@ nml_write_obj (st_parameter_dt *dtp, namelist_info * obj, index_type offset,
 	{
 	  if (rep_ctr > 1)
 	    {
-	      st_sprintf(rep_buff, " %d*", rep_ctr);
+	      sprintf(rep_buff, " %d*", rep_ctr);
 	      write_character (dtp, rep_buff, strlen (rep_buff));
 	      dtp->u.p.no_leading_blank = 1;
 	    }
@@ -1657,11 +985,13 @@ nml_write_obj (st_parameter_dt *dtp, namelist_info * obj, index_type offset,
               break;
 
 	    case GFC_DTYPE_CHARACTER:
-	      if (dtp->u.p.nml_delim)
-		write_character (dtp, &dtp->u.p.nml_delim, 1);
+	      tmp_delim = dtp->u.p.current_unit->flags.delim;
+	      if (dtp->u.p.nml_delim == '"')
+		dtp->u.p.current_unit->flags.delim = DELIM_QUOTE;
+	      if (dtp->u.p.nml_delim == '\'')
+		dtp->u.p.current_unit->flags.delim = DELIM_APOSTROPHE;
 	      write_character (dtp, p, obj->string_length);
-	      if (dtp->u.p.nml_delim)
-		write_character (dtp, &dtp->u.p.nml_delim, 1);
+	      dtp->u.p.current_unit->flags.delim = tmp_delim;
               break;
 
 	    case GFC_DTYPE_REAL:
@@ -1686,32 +1016,43 @@ nml_write_obj (st_parameter_dt *dtp, namelist_info * obj, index_type offset,
 
 	      /* First ext_name => get length of all possible components  */
 
-	      ext_name = (char*)get_mem ( (base_name ? strlen (base_name) : 0)
-					+ (base ? strlen (base->var_name) : 0)
+	      base_name_len = base_name ? strlen (base_name) : 0;
+	      base_var_name_len = base ? strlen (base->var_name) : 0;
+	      ext_name = (char*)get_mem ( base_name_len
+					+ base_var_name_len
 					+ strlen (obj->var_name)
 					+ obj->var_rank * NML_DIGITS
 					+ 1);
 
-	      strcpy(ext_name, base_name ? base_name : "");
-	      clen = base ? strlen (base->var_name) : 0;
-	      strcat (ext_name, obj->var_name + clen);
-
+	      memcpy (ext_name, base_name, base_name_len);
+	      clen = strlen (obj->var_name + base_var_name_len);
+	      memcpy (ext_name + base_name_len, 
+		      obj->var_name + base_var_name_len, clen);
+	      
 	      /* Append the qualifier.  */
 
+	      tot_len = base_name_len + clen;
 	      for (dim_i = 0; dim_i < obj->var_rank; dim_i++)
 		{
-		  strcat (ext_name, dim_i ? "" : "(");
-		  clen = strlen (ext_name);
-		  st_sprintf (ext_name + clen, "%d", (int) obj->ls[dim_i].idx);
-		  strcat (ext_name, (dim_i == obj->var_rank - 1) ? ")" : ",");
+		  if (!dim_i)
+		    {
+		      ext_name[tot_len] = '(';
+		      tot_len++;
+		    }
+		  sprintf (ext_name + tot_len, "%d", (int) obj->ls[dim_i].idx);
+		  tot_len += strlen (ext_name + tot_len);
+		  ext_name[tot_len] = (dim_i == obj->var_rank - 1) ? ')' : ',';
+		  tot_len++;
 		}
 
+	      ext_name[tot_len] = '\0';
+
 	      /* Now obj_name.  */
 
 	      obj_name_len = strlen (obj->var_name) + 1;
 	      obj_name = get_mem (obj_name_len+1);
-	      strcpy (obj_name, obj->var_name);
-	      strcat (obj_name, "%");
+	      memcpy (obj_name, obj->var_name, obj_name_len-1);
+	      memcpy (obj_name + obj_name_len-1, "%", 2);
 
 	      /* Now loop over the components. Update the component pointer
 		 with the return value from nml_write_obj => this loop jumps
@@ -1792,7 +1133,6 @@ namelist_write (st_parameter_dt *dtp)
   /* Set the delimiter for namelist output.  */
 
   tmp_delim = dtp->u.p.current_unit->flags.delim;
-  dtp->u.p.current_unit->flags.delim = DELIM_NONE;
   switch (tmp_delim)
     {
     case (DELIM_QUOTE):
@@ -1808,10 +1148,12 @@ namelist_write (st_parameter_dt *dtp)
       break;
     }
 
+  /* Temporarily disable namelist delimters.  */
+  dtp->u.p.current_unit->flags.delim = DELIM_NONE;
+
   write_character (dtp, "&", 1);
 
   /* Write namelist name in upper case - f95 std.  */
-
   for (i = 0 ;i < dtp->namelist_name_len ;i++ )
     {
       c = toupper (dtp->namelist_name[i]);
@@ -1827,14 +1169,14 @@ namelist_write (st_parameter_dt *dtp)
 	  t1 = nml_write_obj (dtp, t2, dummy_offset, dummy, dummy_name);
 	}
     }
+
 #ifdef HAVE_CRLF
   write_character (dtp, "  /\r\n", 5);
 #else
   write_character (dtp, "  /\n", 4);
 #endif
 
-  /* Recover the original delimiter.  */
-
+  /* Restore the original delimiter.  */
   dtp->u.p.current_unit->flags.delim = tmp_delim;
 }
 
